from worf.serializers import Serializer


class UserSerializer(Serializer):
    def read(self, model):
        return dict(
            username=model.username,
            lastLogin=model.last_login,
            dateJoined=model.date_joined,
            email=model.email,
        )

    def write(self):
        return [
            "username",
            "email",
        ]


class ProfileSerializer(Serializer):
    def read(self, model):
        return dict(
            username=model.user.username,
            email=model.user.email,
            role=dict(name=model.role.name) if model.role else None,
            skills=[t.api() for t in model.ratedskill_set.all()],
            team=dict(name=model.team.name) if model.team else None,
            tags=[t.api() for t in model.tags.all()],
        )

    def write(self):
        return [
            "role",
            "skills",
            "team",
            "tags",
        ]



from datetime import datetime, timedelta

json_type = "application/json"


def test_profile_detail(client, db, profile_url, user):
    response = client.get(profile_url)
    result = response.json()
    assert response.status_code == 200, result
    assert result["username"] == user.username


def test_profile_list(client, db, profile, user):
    response = client.get("/profiles/")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 1
    assert result["profiles"][0]["username"] == user.username


def test_profile_list_array_filter(client, db, profile_factory, tag_factory):
    tag1, tag2, tag3 = tag_factory.create_batch(3)
    profile_factory.create(tags=[tag1])
    profile_factory.create(tags=[tag2])
    profile_factory.create(tags=[tag1, tag2])
    profile_factory.create(tags=[tag3])
    profile_factory.create()
    response = client.get(f"/profiles/?tags={tag1.pk}&tags={tag2.pk}")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 3


def test_profile_update_fk(client, db, profile_url, role, team):
    payload = dict(role=role.pk, team=team.pk)
    response = client.patch(profile_url, payload, content_type=json_type)
    result = response.json()
    assert response.status_code == 200, result
    assert result["role"]["name"] == role.name
    assert result["team"]["name"] == team.name


def test_profile_update_fk_invalid_role(client, db, profile_url, role, team):
    response = client.patch(profile_url, dict(role=123), content_type=json_type)
    result = response.json()
    assert response.status_code == 422, result
    assert result["message"] == "Invalid role"


def test_profile_update_fk_role_is_not_nullable(client, db, profile_url, role, team):
    response = client.patch(profile_url, dict(role=None), content_type=json_type)
    result = response.json()
    assert response.status_code == 422, result
    assert result["message"] == "Invalid role"


def test_profile_update_fk_team_is_nullable(client, db, profile_url, role, team):
    response = client.patch(profile_url, dict(team=None), content_type=json_type)
    result = response.json()
    assert response.status_code == 200, result
    assert result["team"] is None


def test_profile_update_m2m(client, db, profile_url, tag):
    response = client.patch(profile_url, dict(tags=[tag.pk]), content_type=json_type)
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["tags"]) == 1
    assert result["tags"][0]["id"] == tag.pk
    assert result["tags"][0]["name"] == tag.name


def test_profile_update_m2m_can_be_empty(client, db, profile_url, tag):
    response = client.patch(profile_url, dict(tags=[]), content_type=json_type)
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["tags"]) == 0


def test_profile_update_m2m_is_not_nullable(client, db, profile_url, tag):
    response = client.patch(profile_url, dict(tags=None), content_type=json_type)
    result = response.json()
    assert response.status_code == 422, result
    assert "tags accepts an array, got <class 'NoneType'> None" in result["message"]


def test_profile_update_m2m_must_be_pks(client, db, profile_url, tag):
    payload = dict(tags=["invalid"])
    response = client.patch(profile_url, payload, content_type=json_type)
    result = response.json()
    assert response.status_code == 422, result
    assert "Invalid tags" in result["message"]


def test_profile_update_m2m_through(client, db, profile_url, skill):
    payload = dict(skills=[dict(id=skill.pk, rating=4)])
    response = client.patch(profile_url, payload, content_type=json_type)
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["skills"]) == 1
    assert result["skills"][0]["id"] == skill.pk
    assert result["skills"][0]["name"] == skill.name
    assert result["skills"][0]["rating"] == 4


def test_profile_update_m2m_through_can_be_empty(client, db, profile_url, skill):
    response = client.patch(profile_url, dict(skills=[]), content_type=json_type)
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["skills"]) == 0


def test_profile_update_m2m_through_is_not_nullable(client, db, profile_url, skill):
    response = client.patch(profile_url, dict(skills=None), content_type=json_type)
    result = response.json()
    assert response.status_code == 422, result
    assert "skills accepts an array, got <class 'NoneType'> None" in result["message"]


def test_profile_update_m2m_through_must_be_dicts(client, db, profile_url, skill):
    payload = dict(skills=["invalid"])
    response = client.patch(profile_url, payload, content_type=json_type)
    result = response.json()
    assert response.status_code == 422, result
    assert "Invalid skills" == result["message"]


def test_profile_update_m2m_through_ids_must_be_pks(client, db, profile_url, skill):
    payload = dict(skills=[dict(id="invalid")])
    response = client.patch(profile_url, payload, content_type=json_type)
    result = response.json()
    assert response.status_code == 422, result
    assert "Invalid skills" in result["message"]


def test_profile_update_m2m_through_required_fields(client, db, profile_url, skill):
    payload = dict(skills=[dict(id=skill.pk)])
    response = client.patch(profile_url, payload, content_type=json_type)
    result = response.json()
    assert response.status_code == 422, result
    assert "Invalid skills" in result["message"]


def test_user_detail(client, db, user, user_url):
    response = client.get(user_url)
    result = response.json()
    assert response.status_code == 200, result
    assert result["username"] == user.username


def test_user_list(client, db, user):
    response = client.get("/users/")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["users"]) == 1
    assert result["users"][0]["username"] == user.username


def test_user_list_filters(client, db, user_factory):
    january = datetime.fromisoformat("2021-01-01T00:00:00+00:00")
    february = datetime.fromisoformat("2021-02-01T00:00:00+00:00")
    march = datetime.fromisoformat("2021-03-01T00:00:00+00:00")

    user1 = user_factory.create(date_joined=january)
    user2 = user_factory.create(date_joined=february)
    user3 = user_factory.create(email="test3@test.com", date_joined=march)

    # search by email
    response = client.get(f"/users/?email={user1.email}")
    assert len(response.json()["users"]) == 1

    # search by string
    response = client.get("/users/?q=example.com")
    assert len(response.json()["users"]) == 2
    response = client.get(f"/users/?q={user3.email}")
    assert len(response.json()["users"]) == 1

    # search by username -- filter ignored, username is not in filter_fields
    response = client.get(f"/users/?username={user2.username}")
    assert len(response.json()["users"]) == 3

    # search by date joined
    gte, lte = "2021-02-01T00:00:00Z", "2021-02-15T00:00:00Z"
    response = client.get(f"/users/?dateJoined__gte={gte}&date_joined__lte={lte}")
    assert len(response.json()["users"]) == 1
    assert response.json()["users"][0]["username"] == user2.username


def test_user_list_sort_asc(client, db, user_factory):
    user_factory.create(username="a")
    user_factory.create(username="b")
    response = client.get("/users/?sort=id")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["users"]) == 2
    assert result["users"][0]["username"] == "a"
    assert result["users"][1]["username"] == "b"


def test_user_list_sort_desc(client, db, user_factory):
    user_factory.create(username="a")
    user_factory.create(username="b")
    response = client.get("/users/?sort=-id")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["users"]) == 2
    assert result["users"][0]["username"] == "b"
    assert result["users"][1]["username"] == "a"


def test_user_list_multisort(client, now, db, user_factory):
    user_factory.create(username="a", date_joined=now)
    user_factory.create(username="b", date_joined=now - timedelta(hours=1))
    user_factory.create(username="c", date_joined=now)
    user_factory.create(username="d", date_joined=now)
    response = client.get("/users/?sort=dateJoined&sort=-id")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["users"]) == 4
    assert result["users"][0]["username"] == "b"
    assert result["users"][1]["username"] == "d"
    assert result["users"][2]["username"] == "c"
    assert result["users"][3]["username"] == "a"


def test_user_update(client, db, user_url):
    payload = dict(username="testtest", email="something@example.com")
    response = client.patch(user_url, payload, content_type=json_type)
    result = response.json()
    assert response.status_code == 200, result
    assert result["username"] == "testtest"
    assert result["email"] == "something@example.com"



from urllib.parse import urlencode

from django.http import QueryDict

from url_filter.filtersets import ModelFilterSet


def generate_filterset(model):
    return type(
        f"{model.__name__}FilterSet",
        (ModelFilterSet,),
        dict(Meta=type("Meta", (), dict(model=model))),
    )


def apply_filterset(filter_set, queryset, lookup_kwargs):
    data = QueryDict(urlencode(lookup_kwargs))

    return filter_set(data=data, queryset=queryset).filter()



from django.urls import include, path

from tests.views import DummyAPI, ProfileDetail, ProfileList, UserDetail, UserList

urlpatterns = [
    path("", DummyAPI.as_view()),
    path("profiles/", ProfileList.as_view()),
    path("profiles/<str:id>/", ProfileDetail.as_view()),
    path("users/", UserList.as_view()),
    path("users/<str:id>/", UserDetail.as_view()),
    path("<str:id>/", DummyAPI.as_view()),
]



import re
from worf.exceptions import NamingThingsError

lookup_keywords = [
    "exact",
    "iexact",
    "contains",
    "icontains",
    "in",
    "gt",
    "gte",
    "lt",
    "lte",
    "startswith",
    "istartswith",
    "endswith",
    "iendswith",
    "range",
    "date",
    "year",
    "iso_year",
    "month",
    "day",
    "week",
    "week_day",
    "quarter",
    "time",
    "hour",
    "minute",
    "second",
    "isnull",
    "regex",
    "iregex",
]


def snake_to_camel(snake):
    invalid_msg = f"{snake} is not valid snake case. "
    if snake.lower() != snake:
        raise NamingThingsError(invalid_msg + "It has capital letters!")

    components = snake.split("_")
    if not "".join(x for x in components).isalpha():
        raise NamingThingsError(invalid_msg + "It has special chars!")

    return components[0] + "".join(x.title() for x in components[1:])


def camel_to_snake(camel):
    if camel.lower() == camel:
        return camel

    invalid_msg = f"{camel} is not valid camel case. "
    if not clean_lookup_keywords(camel).isalpha():
        raise NamingThingsError(invalid_msg + "It has non alphabetical chars!")

    snake = ""
    last_was_upper = True
    for value in camel:
        if value.isupper():
            if last_was_upper:
                raise NamingThingsError(
                    invalid_msg + "It has multiple upper-case chars sequentially!"
                )
            snake += "_" + value.lower()
        else:
            snake += value

        last_was_upper = value.isupper()
    return snake


def whitespace_to_camel(string):
    pos = string.find(" ")
    if pos == -1:
        return string[:1].lower() + string[1:]

    new_string = string[:pos] + string[pos + 1 :].capitalize()
    return whitespace_to_camel(new_string)


def clean_lookup_keywords(string):
    return re.sub("__(" + "|".join(lookup_keywords) + ")$", "", string)



from django.core.exceptions import ValidationError
from django.contrib.auth.models import User

from worf.permissions import PublicEndpoint
from worf.views import DetailUpdateAPI, ListAPI

from tests.models import DummyModel, Profile
from tests.serializers import ProfileSerializer, UserSerializer


class DummyAPI(DetailUpdateAPI):
    model = DummyModel
    permissions = [PublicEndpoint]

    def validate_phone(self, value):
        try:
            assert value == "(555) 555-5555"
        except AssertionError:
            raise ValidationError("{value} is not a valid phone number")
        return "+5555555555"


class ProfileList(ListAPI):
    model = Profile
    ordering = ["pk"]
    serializer = ProfileSerializer
    permissions = [PublicEndpoint]
    search_fields = {}
    filter_fields = [
        "tags",
    ]


class ProfileDetail(DetailUpdateAPI):
    model = Profile
    serializer = ProfileSerializer
    permissions = [PublicEndpoint]


class UserList(ListAPI):
    model = User
    ordering = ["pk"]
    serializer = UserSerializer
    permissions = [PublicEndpoint]
    search_fields = {
        "or": [
            "email",
            "username",
        ]
    }
    filter_fields = [
        "email",
        "date_joined__gte",
        "date_joined__lte",
    ]
    sort_fields = [
        "id",
        "date_joined",
    ]


class UserDetail(DetailUpdateAPI):
    model = User
    serializer = UserSerializer
    permissions = [PublicEndpoint]



from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db import models
from django.db.utils import IntegrityError

from worf.casing import snake_to_camel
from worf.shortcuts import get_instance_or_http404
from worf.views.base import AbstractBaseAPI


class DetailAPI(AbstractBaseAPI):
    lookup_field = "id"
    lookup_url_kwarg = "id"
    instance = None

    def get(self, request, *args, **kwargs):
        return self.render_to_response()

    def serialize(self):
        """Return the model api, used for responses."""
        serializer = self.get_serializer()
        payload = serializer.read(self.get_instance())
        if not isinstance(payload, dict):
            raise ImproperlyConfigured(f"{serializer} did not return a dictionary")
        return payload

    def get_instance(self):
        # TODO support multiple lookup_fields
        self.lookup_kwargs = {self.lookup_field: self.kwargs[self.lookup_url_kwarg]}

        self.validate_lookup_field_values()

        if self.instance is None:
            self.instance = get_instance_or_http404(self.model, **self.lookup_kwargs)

        return self.instance

    def set_foreign_key(self, instance, key):
        related_model = self.get_related_model(key)
        try:
            related_instance = (
                related_model.objects.get(pk=self.bundle[key])
                if self.bundle[key] is not None
                else None
            )
        except related_model.DoesNotExist as e:
            raise ValidationError(f"Invalid {snake_to_camel(key)}") from e
        setattr(instance, key, related_instance)

    def set_many_to_many(self, instance, key):
        try:
            getattr(instance, key).set(self.bundle[key])
        except (IntegrityError, ValueError) as e:
            raise ValidationError(f"Invalid {snake_to_camel(key)}") from e

    def set_many_to_many_with_through(self, instance, key):
        try:
            attr = getattr(self.model, key)

            through_model = attr.through
            model_name = self.model._meta.model_name
            target_field_name = attr.field.m2m_target_field_name()
            reverse_name = attr.field.m2m_reverse_name()

            getattr(instance, key).clear()

            through_model.objects.bulk_create(
                [
                    through_model(
                        **{
                            key: value
                            for key, value in item.items()
                            if key != target_field_name
                        },
                        **{
                            model_name: instance,
                            reverse_name: item[target_field_name],
                        },
                    )
                    for item in self.bundle[key]
                ]
            )
        except (AttributeError, IntegrityError, ValueError) as e:
            raise ValidationError(f"Invalid {snake_to_camel(key)}") from e

    def update(self):
        instance = self.get_instance()

        self.validate()

        for key in self.bundle.keys():
            attr = getattr(self.model, key)

            if isinstance(attr.field, models.ForeignKey):
                self.set_foreign_key(instance, key)
                continue

            if isinstance(attr.field, models.ManyToManyField):
                if not attr.through._meta.auto_created:
                    self.set_many_to_many_with_through(instance, key)
                    continue

                self.set_many_to_many(instance, key)
                continue

            setattr(instance, key, self.bundle[key])

        instance.save()
        instance.refresh_from_db()

        return instance

    def validate(self):
        for key in self.bundle.keys():
            self.validate_bundle(key)

            field = self.model._meta.get_field(key)

            if self.bundle[key] is None and not field.null:
                raise ValidationError(f"Invalid {snake_to_camel(key)}")


class DetailUpdateAPI(DetailAPI):
    def patch(self, request, *args, **kwargs):
        self.update()
        return self.get(request)



from django.db import models
from django.contrib.auth.models import User


class DummyModel(models.Model):
    id = models.CharField(max_length=64, primary_key=True)
    email = models.CharField(max_length=64)
    phone = models.CharField(max_length=64)

    def api(self):
        return dict(id=self.id, email=self.email, phone=self.phone)

    def api_update_fields(self):
        return [
            "id",
            "email",
            "phone",
        ]


class Role(models.Model):
    name = models.CharField(max_length=64)


class Skill(models.Model):
    name = models.CharField(max_length=200)

    def api(self):
        return dict(
            id=self.pk,
            name=self.name,
        )


class RatedSkill(models.Model):
    class Meta:
        ordering = ["skill__name"]
        unique_together = ["profile", "skill"]

    RATING_CHOICES = [
        (5, "Excellent"),
        (4, "Great"),
        (3, "Average"),
        (2, "Poor"),
        (1, "Bad"),
    ]

    profile = models.ForeignKey("Profile", on_delete=models.CASCADE)
    rating = models.IntegerField(choices=RATING_CHOICES)
    skill = models.ForeignKey("Skill", on_delete=models.CASCADE)

    def __str__(self):
        return f"{self.skill.name}: {self.get_rating_display()}"

    def api(self):
        return dict(
            id=self.skill.pk,
            name=self.skill.name,
            rating=self.rating,
        )


class Tag(models.Model):
    name = models.CharField(max_length=200)

    def api(self):
        return dict(
            id=self.pk,
            name=self.name,
        )


class Team(models.Model):
    name = models.CharField(max_length=64)


class Profile(models.Model):
    user = models.OneToOneField(User, on_delete=models.CASCADE)
    role = models.ForeignKey(Role, on_delete=models.CASCADE)
    team = models.ForeignKey(Team, blank=True, null=True, on_delete=models.SET_NULL)
    skills = models.ManyToManyField(Skill, through=RatedSkill)
    tags = models.ManyToManyField(Tag)
